import warnings
import abc
from collections.abc import Mapping
from pathlib import Path
import yaml


class Config(abc.ABC):
    """An abstract class to create configuration files (stored as YAML)

    Notes
    -----
    For examples, see test_config.py or the concrete classes
    (UserSettings, Internal)
    """

    def __init__(self):
        self._init_values()

        # resolve home directory
        path = self.path()

        if not path.exists():
            defaults = self._get_defaults()
            path.write_text(yaml.dump(defaults))
        else:
            try:
                content = self._load_from_file()
                loaded = True
            except Exception as e:
                warnings.warn(
                    f"Error loading {str(path)!r}: {e}\n\n"
                    "reverting to default values"
                )
                loaded = False
                content = self._get_defaults()

            if loaded and not isinstance(content, Mapping):
                warnings.warn(
                    f"Error loading {str(path)!r}. Expected a dictionary "
                    f"but got {type(content).__name__}, "
                    "reverting to default values"
                )
                content = self._get_defaults()

            self._set_data(content)

    def _load_from_file(self):
        path = self.path()
        config = self.load_config()

        if config:
            content = config
        else:
            # this might happen if using multiprocessing: the first process
            # won't see the file so it'll proceed writing it, but upcoming
            # processes might see an empty file (file has been created but
            # writing hasn't finished). In such case, text will be None. If
            # so, we simply load the default values
            content = self._get_defaults()

        for key, type_ in self.__annotations__.items():
            value = content.get(key, None)

            if value is not None and not isinstance(value, type_):
                default = getattr(self, key)
                warnings.warn(
                    f"Corrupted config file {str(path)!r}: "
                    f"expected {key!r} to contain an object "
                    f"with type {type_.__name__}, but got "
                    f"{type(value).__name__}. Reverting to "
                    f"default value {default}"
                )
                content[key] = default

        return content

    def _get_defaults(self):
        """Extract values from the annotations and return a dictionary"""
        return {key: getattr(self, key) for key in self.__annotations__}

    def _set_data(self, data):
        """Take a dictionary and store it in the annotations"""
        for key in self.__annotations__:
            if key in data:
                setattr(self, key, data[key])

    def _init_values(self):
        """
        Iterate over annotations to initialize values. This is only relevant
        when any of the annotations has a factory method to initialize the
        values. If they value is a literal, no changes happen.
        """
        for key in self.__annotations__:
            name = f"{key}_default"

            # if there is a method with such name, call it and store the output
            if hasattr(self, name):
                value = getattr(self, name)()
                # call __setattr__ on the superclass so we skip the part
                # where we overwrite the YAML file, here we only want to
                # set the default values
                super().__setattr__(key, value)

    def _write(self):
        """Writes data to the YAML file"""
        data = self._get_defaults()
        self.path().write_text(yaml.dump(data))

    def __setattr__(self, name, value):
        if name not in self.__annotations__:
            raise ValueError(f"{name} not a valid field")
        else:
            super().__setattr__(name, value)
            self._write()

    def load_config(self):
        config = None
        path = self.path()
        is_config_exist = Path.is_file(path)
        if is_config_exist:
            text = path.read_text()
            if text:
                config = yaml.safe_load(text)

        return config

    @abc.abstractclassmethod
    def path(cls):
        """Returns the path to the YAML file"""
        pass



import difflib
from ploomber_core.exceptions import ValidationError
from ploomber_core.io import pretty_print
import collections.abc


def _is_set_or_sequence(obj) -> bool:
    """
    Checks if obj is a sequence or a set.
    Returns False if obj is a string.
    """
    if isinstance(obj, str):
        return False

    return isinstance(obj, (collections.abc.Sequence, set))


def keys(valid, passed, name="spec", show_matches=True):
    """
    Checks if given values are valid. If not, raises an error with
    suggested valid values.

    Parameters
    ----------
    valid : list
        A list of valid values

    passed : str, list, tuple, or a set
        The value to check

    required : list
        Required values

    name : str, default 'spec'
        The name of the parameter

    show_matches : bool, default True
        Show suggested values in the error message

    Raises
    ------
    ValidationError if passed elements are not in valid
    """
    is_set_or_sequence = _is_set_or_sequence(passed)

    if not is_set_or_sequence or passed is None:
        passed = [passed]

    passed = set(passed)

    if valid:
        extra = passed - set(valid)

        if extra:
            err_message = f"Error validating argument '{name}', "

            if len(extra) < 2:
                err_message += "the following value isn't valid: "
            else:
                err_message += "the following values aren't valid: "

            err_message += f"{pretty_print.iterable(extra)}. "

            if len(valid) > 0:
                err_message += f"Valid values are: {pretty_print.iterable(valid)}. "

            if show_matches:
                input = str(passed.pop()) if len(passed) == 1 else passed
                input_matches = get_formatted_close_matches(input, valid)
                if len(input_matches) > 0:
                    err_message += f"\nDid you mean {input_matches}?"

            raise ValidationError(err_message)


def get_formatted_close_matches(input, valid) -> str:
    """
    Returns a formatted string of the best "good enough" matches
    """
    close_matches = []

    valid = list(str(i) for i in valid)

    if len(valid) > 0 and len(input) > 0:
        if _is_set_or_sequence(input):
            for i in input:
                if not isinstance(i, str):
                    i = str(i)
                input_matches = difflib.get_close_matches(i, valid)
                close_matches += input_matches
        else:
            close_matches = difflib.get_close_matches(str(input), valid)

        # remove dups
        close_matches = list(set(close_matches))
        close_matches = pretty_print.iterable(close_matches, last_delimiter="or")

    return close_matches



from functools import wraps
import importlib


def requires(pkgs, name=None, extra_msg=None, pip_names=None):
    """
    Decorator to check if packages were imported, raise ModuleNotFoundError with an
    appropriate message for missing ones

    Error message:
    a, b are required to use function. Install them by running pip install a b

    Parameters
    ----------
    pkgs : list
        The names of the packages required

    name
        The name of the module/function/class to show in the error message,
        if None, the decorated function __name__ attribute is used

    extra_msg
        Append this extra message to the end

    pip_names : list
        Pip package names to show in the suggested "pip install {name}"
        command, use it if different to the package name itself

    """

    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            check_installed(
                pkgs=pkgs,
                name=name or f.__name__,
                extra_msg=extra_msg,
                pip_names=pip_names,
            )
            return f(*args, **kwargs)

        return wrapper

    return decorator


def check_installed(pkgs, name, extra_msg=None, pip_names=None):
    """
    A function to check if packages were imported, raise ModuleNotFoundError with an
    appropriate message for missing ones

    Error message:
    a, b are required to use function. Install them by running pip install a b

    Parameters
    ----------
    pkgs : list
        The names of the packages required

    name
        The name of the module/function/class to show in the error message,
        if None, the decorated function __name__ attribute is used

    extra_msg
        Append this extra message to the end

    pip_names : list
        Pip package names to show in the suggested "pip install {name}"
        command, use it if different to the package name itself

    """
    pkgs = [pkg.replace("-", "_") for pkg in pkgs]

    is_pkg_missing = [importlib.util.find_spec(pkg) is None for pkg in pkgs]

    if any(is_pkg_missing):
        missing_pkgs = [
            name
            for name, is_missing in zip(pip_names or pkgs, is_pkg_missing)
            if is_missing
        ]

        raise ModuleNotFoundError(
            _make_requires_error_message(missing_pkgs, name, extra_msg)
        )


def _make_requires_error_message(missing_pkgs, fn_name, extra_msg):
    names_str = " ".join(repr(pkg) for pkg in missing_pkgs)

    error_msg = "{} {} required to use {}. Install with: " "pip install {}".format(
        names_str,
        "is" if len(missing_pkgs) == 1 else "are",
        repr(fn_name),
        names_str,
    )

    if extra_msg:
        error_msg += "\n" + extra_msg

    return error_msg



def iterable(obj, delimiter=",", last_delimiter="and", repr_=False) -> str:
    """
    Returns a formatted string representation of an array
    """
    if repr_:
        sorted_ = sorted(repr(element) for element in obj)
    else:
        sorted_ = sorted(f"'{element}'" for element in obj)

    if len(sorted_) > 1:
        sorted_[-1] = f"{last_delimiter} {sorted_[-1]}"

    return f"{delimiter} ".join(sorted_)



import warnings
from functools import wraps

from ploomber_core.warnings import PloomberDeprecationWarning

COMMUNITY = "If you need help migrating, contact us: https://ploomber.io/community"


def parameter_renamed(
    deprecated_in,
    remove_in,
    *,
    old_name,
    old_value,
    new_name,
    new_value,
    custom_message=None,
):
    """

    Notes
    -----
    .. versionadded:: 0.1
    """

    using_old_parameter = old_value != "deprecated"

    if using_old_parameter:
        warnings.warn(
            f"{old_name!r} was renamed to {new_name!r} in version "
            f"{deprecated_in}. {old_name!r} will be removed in "
            f"{remove_in}. {_message_end(custom_message)}",
            PloomberDeprecationWarning,
        )

    return old_value if using_old_parameter else new_value


def parameter_deprecated(deprecated_in, remove_in, *, name_old, value_passed):
    """

    Notes
    -----
    .. versionadded:: 0.1
    """
    warn = value_passed != "deprecated"

    if warn:
        warnings.warn(
            f"{name_old!r} was deprecated in version "
            f"{deprecated_in}. {name_old!r} will be removed "
            f"in {remove_in}. {COMMUNITY}",
            PloomberDeprecationWarning,
        )

    return warn


def parameter_default_changed(
    changed_in,
    *,
    name,
    old_default,
    new_default,
    value,
):
    """

    Notes
    -----
    .. versionadded:: 0.1
    """
    using_default = value == "warn"

    if using_default:
        warnings.warn(
            f"The default value of {name} will change from {old_default!r} to "
            f"{new_default!r} in {changed_in}.",
            PloomberDeprecationWarning,
        )

    return old_default if using_default else value


def _message_end(custom_message):
    if custom_message:
        return f"{custom_message}. {COMMUNITY}"
    else:
        return COMMUNITY


def function(deprecated_in, remove_in, *, new_name=None, custom_message=None):
    """A decorator for deprecated functions

    Notes
    -----
    .. versionadded:: 0.1
    """

    def decorator(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            what = "deprecated" if not new_name else f"renamed to {new_name!r}"
            name_old = fn.__name__
            warnings.warn(
                f"Function {name_old!r} was "
                f"{what} in version {deprecated_in}. {name_old!r} will be "
                f"removed in version {remove_in}. {_message_end(custom_message)}",
                category=PloomberDeprecationWarning,
            )
            return fn(*args, **kwargs)

        return wrapper

    return decorator


def method(deprecated_in, remove_in, *, new_name=None, custom_message=None):
    """A decorator for deprecated methods

    Notes
    -----
    .. versionchanged:: 0.1
        Added ``new_name`` and ``custom_message``
    """

    def decorator(fn):
        @wraps(fn)
        def wrapper(self, *args, **kwargs):
            what = "deprecated" if not new_name else f"renamed to {new_name!r}"
            name_old = fn.__name__
            warnings.warn(
                f"{name_old!r} from {type(self).__name__!r} was "
                f"{what} in version {deprecated_in}. {name_old!r} will be "
                f"removed in version {remove_in}. {_message_end(custom_message)}",
                category=PloomberDeprecationWarning,
            )
            return fn(self, *args, **kwargs)

        return wrapper

    return decorator



from typing import Any, Optional


# Validate the input of type string
def str_param(item: Any, name: str) -> str:
    if not isinstance(item, str):
        raise TypeError(
            f"TypeError: Variable not supported/wrong type: "
            f"{item}, {name}, should be a str"
        )

    return item


def opt_str_param(name: str, item: Any) -> Optional[str]:
    # Can leverage regular string function
    if item is not None and not isinstance(item, str):
        raise TypeError(
            f"TypeError: Variable not supported/wrong type: "
            f"{item}, {name}, should be a str"
        )
    return item



import re
import ast
from glob import glob
from os.path import basename, splitext

from setuptools import find_packages
from setuptools import setup

_version_re = re.compile(r"__version__\s+=\s+(.*)")

with open("src/ploomber_core/__init__.py", "rb") as f:
    VERSION = str(
        ast.literal_eval(_version_re.search(f.read().decode("utf-8")).group(1))
    )

REQUIRES = [
    "click",
    "pyyaml",
    "posthog",
]

DEV = [
    "pytest",
    "flake8",
    "twine",
    "invoke",
    "pkgmt",
]

setup(
    name="ploomber-core",
    version=VERSION,
    description=None,
    license=None,
    author=None,
    author_email=None,
    url=None,
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    include_package_data=True,
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
    ],
    keywords=[],
    install_requires=REQUIRES,
    extras_require={
        "dev": DEV,
    },
    entry_points={},
)



import typing as t
from functools import wraps
from gettext import gettext as _

from click.exceptions import ClickException
from click._compat import get_text_stderr
from click.utils import echo

COMMUNITY_LINK = "https://ploomber.io/community"

COMMUNITY = (
    "\nIf you need help solving this " "issue, send us a message: " + COMMUNITY_LINK
)


def get_community_link():
    return COMMUNITY_LINK


def _format_message(exception):
    if hasattr(exception, "format_message"):
        return exception.format_message()
    else:
        return str(exception)


def _build_message(exception):
    msg = _format_message(exception)

    while exception.__cause__:
        msg += f"\n{_format_message(exception.__cause__)}"
        exception = exception.__cause__

    return msg


class BaseException(ClickException):
    """
    A subclass of ClickException that adds support for printing error messages
    from chained exceptions
    """

    def __init__(self, message, type_=None):
        super().__init__(message)
        self.type_ = type_

    def get_message(self):
        return f"Error: {_build_message(self)}"

    def show(self, file: t.Optional[t.IO] = None) -> None:
        if file is None:
            file = get_text_stderr()

        echo(_(self.get_message()), file=file)


class PloomberValueError(ValueError):
    """Subclass of ValueError that displays the community link"""

    def __init__(self, message):
        super().__init__(f"{message}. {COMMUNITY}")


class PloomberTypeError(TypeError):
    """Subclass of TypeError that displays the community link"""

    def __init__(self, message):
        super().__init__(f"{message}. {COMMUNITY}")


class PloomberKeyError(KeyError):
    """Subclass of KeyError that displays the community link"""

    def __init__(self, message):
        super().__init__(f"{message}. {COMMUNITY.strip()}")


class ValidationError(BaseException):
    """Raised when failed to validate input data"""

    pass


def _add_community_link(e):
    if COMMUNITY not in e.args[0]:
        message = e.args[0] + COMMUNITY
        e.args = (message,)

    return e


def modify_exceptions(fn):
    """
    A decorator that catches ValueError, TypeError, ModifiableException and modifies
    the original error message

    Notes
    -----
    .. versionadded:: 0.1.1

    .. versionchanged:: 0.2.10
        Decorator also modifies exceptions with if they have a "modify_exception"
        attribute and it's set to True
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except (ValueError, TypeError) as e:
            _add_community_link(e)
            raise

        except Exception as e:
            if getattr(e, "modify_exception", False):
                _add_community_link(e)
            raise

    return wrapper
